import { h } from '@phosphor/virtualdom';
import URI from '@wiptheia/core/lib/common/uri';
import { MessageService, ResourceProvider, CommandService, MenuPath } from '@wiptheia/core';
import { ContextMenuRenderer, VirtualWidget, LabelProvider, StatefulWidget } from '@wiptheia/core/lib/browser';
import { EditorManager, EditorWidget, EditorOpenerOptions } from '@wiptheia/editor/lib/browser';
import { WorkspaceService } from '@wiptheia/workspace/lib/browser';
import { Git, GitFileChange, GitFileStatus, Repository, WorkingDirectoryStatus, CommitWithChanges } from '../common';
import { GitWatcher } from '../common/git-watcher';
import { GitRepositoryProvider } from './git-repository-provider';
import { GitCommitMessageValidator } from './git-commit-message-validator';
import { GitAvatarService } from './history/git-avatar-service';
export interface GitFileChangeNode extends GitFileChange {
    readonly icon: string;
    readonly label: string;
    readonly description: string;
    readonly caption?: string;
    readonly extraIconClassName?: string;
    readonly commitSha?: string;
    selected?: boolean;
}
export declare namespace GitFileChangeNode {
    function is(node: Object | undefined): node is GitFileChangeNode;
}
export declare class GitWidget extends VirtualWidget implements StatefulWidget {
    protected readonly git: Git;
    protected readonly gitWatcher: GitWatcher;
    protected readonly contextMenuRenderer: ContextMenuRenderer;
    protected readonly resourceProvider: ResourceProvider;
    protected readonly messageService: MessageService;
    protected readonly commandService: CommandService;
    protected readonly repositoryProvider: GitRepositoryProvider;
    protected readonly labelProvider: LabelProvider;
    protected readonly workspaceService: WorkspaceService;
    protected readonly avatarService: GitAvatarService;
    protected readonly commitMessageValidator: GitCommitMessageValidator;
    private static MESSAGE_BOX_MIN_HEIGHT;
    protected stagedChanges: GitFileChangeNode[];
    protected unstagedChanges: GitFileChangeNode[];
    protected mergeChanges: GitFileChangeNode[];
    protected message: string;
    protected messageBoxHeight: number;
    protected status: WorkingDirectoryStatus | undefined;
    protected scrollContainer: string;
    protected commitMessageValidationResult: GitCommitMessageValidator.Result | undefined;
    protected lastCommit: {
        commit: CommitWithChanges;
        avatar: string;
    } | undefined;
    protected lastHead: string | undefined;
    protected readonly editorManager: EditorManager;
    constructor(git: Git, gitWatcher: GitWatcher, contextMenuRenderer: ContextMenuRenderer, resourceProvider: ResourceProvider, messageService: MessageService, commandService: CommandService, repositoryProvider: GitRepositoryProvider, labelProvider: LabelProvider, workspaceService: WorkspaceService, avatarService: GitAvatarService, commitMessageValidator: GitCommitMessageValidator);
    protected init(): void;
    initialize(repository: Repository | undefined): Promise<void>;
    storeState(): object;
    restoreState(oldState: any): void;
    protected undo(): Promise<void>;
    commit(repository?: Repository, options?: 'amend' | 'sign-off', message?: string): Promise<void>;
    protected updateView(status: WorkingDirectoryStatus | undefined): Promise<void>;
    protected renderCommitMessage(): h.Child;
    protected onCommitMessageChange(e: Event): void;
    protected resize(textArea: HTMLTextAreaElement): void;
    protected validateCommitMessage(input: string | undefined): Promise<GitCommitMessageValidator.Result | undefined>;
    protected render(): h.Child;
    protected createChildContainer(): HTMLElement;
    protected getLastCommit(): Promise<{
        commit: CommitWithChanges;
        avatar: string;
    } | undefined>;
    protected renderLastCommit(): h.Child;
    protected renderCommandBar(repository: Repository | undefined): h.Child;
    protected getUserConfig(repository: Repository): Promise<[string, string]>;
    protected renderGitItemButtons(repository: Repository, change: GitFileChange): h.Child;
    protected renderGitItem(repository: Repository | undefined, change: GitFileChangeNode): h.Child;
    protected renderChangesHeader(title: string): h.Child;
    protected renderMergeChanges(repository: Repository | undefined): h.Child | undefined;
    protected renderStagedChanges(repository: Repository | undefined): h.Child | undefined;
    protected renderUnstagedChanges(repository: Repository | undefined): h.Child | undefined;
    protected logError(error: any): void;
    protected getStatusCaption(status: GitFileStatus, staged?: boolean): string;
    protected getAbbreviatedStatusCaption(status: GitFileStatus, staged?: boolean): string;
    findChange(uri: URI): GitFileChange | undefined;
    openChange(change: GitFileChange, options?: EditorOpenerOptions): Promise<EditorWidget | undefined>;
    protected createChangeUri(change: GitFileChange): URI;
    protected resetCommitMessages(): void;
}
export declare namespace GitWidget {
    namespace ContextMenu {
        const PATH: MenuPath;
        const OTHER_GROUP: MenuPath;
        const COMMIT_GROUP: MenuPath;
    }
    namespace Styles {
        const MAIN_CONTAINER = "theia-git-main-container";
        const COMMIT_MESSAGE_CONTAINER = "theia-git-commit-message-container";
        const COMMIT_MESSAGE = "theia-git-commit-message";
        const VALIDATION_MESSAGE = "theia-git-commit-validation-message";
        const LAST_COMMIT_CONTAINER = "theia-git-last-commit-container";
        const LAST_COMMIT_DETAILS = "theia-git-last-commit-details";
        const LAST_COMMIT_MESSAGE_AVATAR = "theia-git-last-commit-message-avatar";
        const LAST_COMMIT_MESSAGE_SUMMARY = "theia-git-last-commit-message-summary";
        const LAST_COMMIT_MESSAGE_TIME = "theia-git-last-commit-message-time";
        const FLEX_CENTER = "flex-container-center";
        const NO_SELECT = "no-select";
    }
}
