/*
 * Copyright (C) 2018 TypeFox and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
 */

import { inject, injectable } from 'inversify';
import URI from '@wiptheia/core/lib/common/uri';
import { ILogger } from '@wiptheia/core/lib/common/logger';
import { Event, Emitter } from '@wiptheia/core/lib/common/event';
import { Tree } from '@wiptheia/core/lib/browser/tree/tree';
import { DepthFirstTreeIterator } from '@wiptheia/core/lib/browser/tree/tree-iterator';
import { DisposableCollection } from '@wiptheia/core/lib/common/disposable';
import { PreferenceChangeEvent } from '@wiptheia/core/lib/browser/preferences/preference-proxy';
import { TreeDecorator, TreeDecoration } from '@wiptheia/core/lib/browser/tree/tree-decorator';
import { Git } from '../common/git';
import { GitWatcher } from '../common/git-watcher';
import { WorkingDirectoryStatus } from '../common/git-model';
import { GitRepositoryProvider } from './git-repository-provider';
import { GitFileChange, GitFileStatus } from '../common/git-model';
import { GitPreferences, GitConfiguration } from './git-preferences';

@injectable()
export class GitDecorator implements TreeDecorator {

    readonly id = 'theia-git-decorator';

    protected readonly toDisposeOnRepositoryChange: DisposableCollection;
    protected readonly emitter: Emitter<(tree: Tree) => Map<string, TreeDecoration.Data>>;

    protected enabled: boolean;
    protected showColors: boolean;

    constructor(
        @inject(Git) protected readonly git: Git,
        @inject(GitRepositoryProvider) protected readonly repositoryProvider: GitRepositoryProvider,
        @inject(GitWatcher) protected readonly watcher: GitWatcher,
        @inject(GitPreferences) protected readonly preferences: GitPreferences,
        @inject(ILogger) protected readonly logger: ILogger) {
        this.emitter = new Emitter();
        this.toDisposeOnRepositoryChange = new DisposableCollection();
        this.repositoryProvider.onDidChangeRepository(async repository => {
            this.toDisposeOnRepositoryChange.dispose();
            if (repository) {
                this.toDisposeOnRepositoryChange.pushAll([
                    await this.watcher.watchGitChanges(repository),
                    this.watcher.onGitEvent(event => this.fireDidChangeDecorations((tree: Tree) => this.collectDecorators(tree, event.status)))
                ]);
            }
        });
        this.preferences.onPreferenceChanged(event => this.handlePreferenceChange(event));
        this.enabled = this.preferences['git.decorations.enabled'];
        this.showColors = this.preferences['git.decorations.colors'];
    }

    async decorations(tree: Tree): Promise<Map<string, TreeDecoration.Data>> {
        const { selectedRepository } = this.repositoryProvider;
        if (selectedRepository) {
            const status = await this.git.status(selectedRepository);
            return this.collectDecorators(tree, status);
        }
        return new Map();
    }

    get onDidChangeDecorations(): Event<(tree: Tree) => Map<string, TreeDecoration.Data>> {
        return this.emitter.event;
    }

    protected fireDidChangeDecorations(event: (tree: Tree) => Map<string, TreeDecoration.Data>): void {
        this.emitter.fire(event);
    }

    protected collectDecorators(tree: Tree, status: WorkingDirectoryStatus): Map<string, TreeDecoration.Data> {
        const result = new Map();
        if (tree.root === undefined || !this.enabled) {
            return result;
        }
        const markers = this.appendContainerChanges(tree, status.changes);
        for (const { id } of new DepthFirstTreeIterator(tree.root)) {
            const marker = markers.get(id);
            if (marker) {
                result.set(id, marker);
            }
        }
        return new Map(Array.from(result.values()).map(m => [m.uri, this.toDecorator(m)] as [string, TreeDecoration.Data]));
    }

    protected appendContainerChanges(tree: Tree, changes: GitFileChange[]): Map<string, GitFileChange> {
        const result: Map<string, GitFileChange> = new Map();
        // We traverse up and assign the highest Git file change status the container directory.
        // Note, instead of stopping at the WS root, we traverse up the driver root.
        // We will filter them later based on the expansion state of the tree.
        for (const [uri, change] of new Map(changes.map(m => [new URI(m.uri), m] as [URI, GitFileChange])).entries()) {
            const uriString = uri.toString();
            result.set(uriString, change);
            let parentUri: URI | undefined = uri.parent;
            while (parentUri && !parentUri.path.isRoot) {
                const parentUriString = parentUri.toString();
                const existing = result.get(parentUriString);
                if (existing === undefined || this.compare(existing, change) < 0) {
                    result.set(parentUriString, {
                        uri: parentUriString,
                        status: change.status,
                        staged: !!change.staged
                    });
                    parentUri = parentUri.parent;
                } else {
                    parentUri = undefined;
                }
            }
        }
        return result;
    }

    protected toDecorator(change: GitFileChange): TreeDecoration.Data {
        const data = GitFileStatus.toAbbreviation(change.status, change.staged);
        const color = this.getDecorationColor(change.status, change.staged);
        const tooltip = GitFileStatus.toString(change.status, change.staged);
        let decorationData: TreeDecoration.Data = {
            tailDecorations: [
                {
                    data,
                    fontData: {
                        color
                    },
                    tooltip
                }
            ]
        };
        if (this.showColors) {
            decorationData = {
                ...decorationData,
                fontData: {
                    color
                }
            };
        }
        return decorationData;
    }

    protected compare(left: GitFileChange, right: GitFileChange): number {
        return GitFileStatus.statusCompare(left.status, right.status);
    }

    protected getDecorationColor(status: GitFileStatus, staged?: boolean): string {
        switch (status) {
            case GitFileStatus.New: return 'var(--theia-success-color0)';
            case GitFileStatus.Renamed: // Fall through.
            case GitFileStatus.Copied: // Fall through.
            case GitFileStatus.Modified: return 'var(--theia-brand-color0)';
            case GitFileStatus.Deleted: return 'var(--theia-warn-color0)';
            case GitFileStatus.Conflicted: return 'var(--theia-error-color0)';
        }
    }

    protected async handlePreferenceChange(event: PreferenceChangeEvent<GitConfiguration>): Promise<void> {
        let refresh = false;
        const { preferenceName, newValue } = event;
        if (preferenceName === 'git.decorations.enabled') {
            const enabled = !!newValue;
            if (this.enabled !== enabled) {
                this.enabled = enabled;
                refresh = true;
            }
        }
        if (preferenceName === 'git.decorations.colors') {
            const showColors = !!newValue;
            if (this.showColors !== showColors) {
                this.showColors = showColors;
                refresh = true;
            }
        }
        const repository = this.repositoryProvider.selectedRepository;
        if (refresh && repository) {
            const status = await this.git.status(repository);
            this.fireDidChangeDecorations((tree: Tree) => this.collectDecorators(tree, status));
        }
    }

}
